@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2019 morguldir
+# Copyright (C) 2019-2021 morguldir
 # Copyright (C) 2014 Thomas Amland
 #
 # This program is free software: you can redistribute it and/or modify
@@ -16,15 +16,19 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import print_function
 from __future__ import unicode_literals
 from collections import namedtuple
 from enum import Enum
 
+import concurrent.futures
 import datetime
 import json
 import logging
 import requests
 import base64
+import time
+import uuid
 from .models import Artist, Album, Track, Video, Playlist, SearchResult, Category, Role
 
 try:
@@ -47,6 +51,26 @@ class VideoQuality(Enum):
     low = 'LOW'
 
 
+class LinkLogin(object):
+    """
+    The data required for logging in to TIDAL using a remote link, json is the data returned from TIDAL
+    """
+    #: Amount of seconds until the code expires
+    expires_in = None
+    #: The code the user should enter at the uri
+    user_code = None
+    #: The link the user has to visit
+    verification_uri = None
+    #: The link the user has to visit, with the code already included
+    verification_uri_complete = None
+
+    def __init__(self, json):
+        self.expires_in = json['expiresIn']
+        self.user_code = json['userCode']
+        self.verification_uri = json['verificationUri']
+        self.verification_uri_complete = json['verificationUriComplete']
+
+
 class Config(object):
     def __init__(self, quality=Quality.high, video_quality=VideoQuality.high):
         self.quality = quality.value
@@ -62,15 +86,57 @@ class Config(object):
             decode("".join(map(chr, [117, 116, 70, 95, 56]))))
         token = self.api_token
         self.api_token = list((base64.b64decode("d3RjaThkamFfbHlhQnBKaWQuMkMwb3puT2ZtaXhnMA==").decode()))
-        for B in token:
-            self.api_token.remove(B)
+        tok = "".join(([chr(ord(x)+1) for x in token[-6:]]))
+        token2 = token
+        token = token[:9]
+        token += tok
+        tok2 = "".join(([chr(ord(x)+2) for x in token[:-7]]))
+        token = token[8:]
+        token = tok2 + token
+        self.api_token = list((base64.b64decode("UHZ6a2RmMGNFbnhjTnJWa0gua0R5cFpvTGdpalloTg==").decode()))
+        for word in token:
+            self.api_token.remove(word)
         self.api_token = "".join(self.api_token)
+        string = ""
+        save = False
+        if not isinstance(token2, str):
+            save = True
+            string = "".encode('ISO-8859-1')
+            token2 = token2.encode('ISO-8859-1')
+        tok = string.join(([chr(ord(x) + 20) for x in token2[:-7]]))
+        token2 = token2[8:]
+        token2 = tok + token2
+        tok2 = string.join(([chr(ord(x)+22) for x in token2[-6:]]))
+        token2 = token2[:9]
+        token2 += tok2
+        self.client_id = list((base64.b64decode("V4g3fVh4NnVVgHZ1QoRhfWgub1krhViET3xpfzF9TVVl"
+                                                "Q1g2ZXd2MnpUZFNPVjNZN3FDM3AzNjc1ST0=").decode('ISO-8859-1')))
+        if save:
+            token2.decode('ISO-8859-1').encode('utf-16')
+            self.client_id = [x.encode('ISO-8859-1') for x in self.client_id]
+        for word in token2:
+            self.client_id.remove(word)
+        self.client_id = "".join(self.client_id)
+        self.client_secret = self.client_id
+        self.client_id = self.api_token
+
 
 class Session(object):
+    #: The TIDAL access token, this is what you use with load_oauth_session
+    access_token = None
+    #: A :class:`datetime` object containing the date the access token will expire
+    expiry_time = None
+    #: A refresh token for retrieving a new access token through refresh_token
+    refresh_token = None
+    #: The type of access token, e.g. Bearer
+    token_type = None
+    #: The id for a TIDAL session, you also need this to use load_oauth_session
+    session_id = None
+    country_code = None
+    #: A :class:`.User` object containing the currently logged in user.
+    user = None
+
     def __init__(self, config=Config()):
-        self.session_id = None
-        self.country_code = None
-        self.user = None
         self._config = config
         """:type _config: :class:`Config`"""
 
@@ -84,6 +150,40 @@ class Session(object):
         self.country_code = country_code
         self.user = User(self, id=user_id)
 
+    def load_oauth_session(self, session_id, token_type, access_token, refresh_token=None):
+        """
+        Login to TIDAL using details from a previous OAuth login, automatically
+        refreshes expired access tokens if refresh_token is supplied as well.
+
+        :param token_type: The type of token, e.g. Bearer
+        :param session_id: The TIDAL session id, has to be a UUID
+        :param access_token: The access token received from an oauth login or refresh
+        :param refresh_token: (Optional) A refresh token that lets you get a new access token after it has expired
+        :return: True if we believe the log in was successful, otherwise false.
+        """
+        try:
+            uuid.UUID(session_id)
+        except ValueError:
+            log.error("Session id did not have a valid UUID format")
+            return False
+        self.session_id = session_id
+        self.token_type = token_type
+        self.access_token = access_token
+        self.refresh_token = refresh_token
+
+        request = self.basic_request('GET', 'sessions')
+        json = request.json()
+        if not request.ok and json['userMessage'].startswith("The token has expired.") and refresh_token:
+            log.debug("The access token has expired, trying to refresh it.")
+            refreshed = self.token_refresh(refresh_token)
+            if not refreshed:
+                return False
+
+        self.country_code = json['countryCode']
+        self.user = User(self, id=json['userId'])
+
+        return True
+
     def login(self, username, password):
         url = urljoin(self._config.api_location, 'login/username')
         headers = {"X-Tidal-Token": self._config.api_token}
@@ -103,14 +203,114 @@ class Session(object):
         self.user = User(self, id=body['userId'])
         return True
 
+    def login_oauth_simple(self, function=print):
+        """
+        Login to TIDAL using a remote link. You can select what function you want to use to display the link
+
+        :param function: The function you want to display the link with
+        :raises: TimeoutError: If the login takes too long
+        """
+        login, future = self.login_oauth()
+        text = "Visit {0} to log in, the code will expire in {1} seconds"
+        function(text.format(login.verification_uri_complete, login.expires_in))
+        future.result()
+
+    def login_oauth(self):
+        """
+        Login to TIDAL with a remote link for limited input devices. The function will return everything you
+        need to log in through a web browser, and will return an future that will run until login.
+        :return: A :class:`LinkLogin` object containing all the data needed to log in remotely, and
+            a :class:`concurrent.futures.Future` that will poll until the login is completed, or until the link expires.
+        :raises: TimeoutError: If the login takes too long
+        """
+        login, future = self._login_with_link()
+        return login, future
+
+    def _login_with_link(self):
+        url = 'https://auth.tidal.com/v1/oauth2/device_authorization'
+        params = {
+            'client_id': self._config.client_id,
+            'scope': 'r_usr w_usr w_sub'
+        }
+
+        request = requests.post(url, params)
+
+        if not request.ok:
+            log.error("Login failed: %s", request.text)
+            request.raise_for_status()
+
+        json = request.json()
+        executor = concurrent.futures.ThreadPoolExecutor()
+        return LinkLogin(json), executor.submit(self._process_link_login, json)
+
+    def _process_link_login(self, json):
+        json = self._wait_for_link_login(json)
+        self.access_token = json['access_token']
+        self.expiry_time = datetime.datetime.now() + datetime.timedelta(seconds=json['expires_in'])
+        self.refresh_token = json['refresh_token']
+        self.token_type = json['token_type']
+        session = self.request('GET', 'sessions')
+        json = session.json()
+        self.session_id = json['sessionId']
+        self.country_code = json['countryCode']
+        self.user = User(self, id=json['userId'])
+
+    def _wait_for_link_login(self, json):
+        expiry = json['expiresIn']
+        interval = json['interval']
+        device_code = json['deviceCode']
+        url = 'https://auth.tidal.com/v1/oauth2/token'
+        params = {
+            'client_id': self._config.client_id,
+            'client_secret': self._config.client_secret,
+            'device_code': device_code,
+            'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
+            'scope': 'r_usr w_usr w_sub'
+        }
+        while expiry > 0:
+            request = requests.post(url, params)
+            json = request.json()
+            if request.ok:
+                return json
+            # Because the requests take time, the expiry variable won't be accurate, so stop if TIDAL says it's expired
+            if json['error'] == 'expired_token':
+                break
+            time.sleep(interval)
+            expiry = expiry - interval
+
+        raise TimeoutError('You took too long to log in')
+
+    def token_refresh(self, refresh_token):
+        """
+        Retrieves a new access token using the specified refresh token, updating the current access token
+        :param refresh_token: The refresh token retrieved when using the OAuth login.
+        :return: True if we believe the token was successfully refreshed, otherwise False
+        """
+        url = 'https://auth.tidal.com/v1/oauth2/token'
+        params = {
+            'grant_type': 'refresh_token',
+            'refresh_token': refresh_token,
+            'client_id': self._config.client_id,
+            'client_secret': self._config.client_id
+        }
+
+        request = requests.post(url, params)
+        json = request.json()
+        if not request.ok:
+            log.debug("The refresh token has expired, a new login is required.")
+            return False
+        self.access_token = json['access_token']
+        self.expiry_time = datetime.datetime.now() + datetime.timedelta(seconds=json['expires_in'])
+        self.token_type = json['token_type']
+        return True
+
     def check_login(self):
         """ Returns true if current session is valid, false otherwise. """
         if self.user is None or not self.user.id or not self.session_id:
             return False
-        url = urljoin(self._config.api_location, 'users/%s/subscription' % self.user.id)
-        return requests.get(url, params={'sessionId': self.session_id}).ok
+        return self.basic_request('GET', 'users/%s/subscription' % self.user.id).ok
 
-    def request(self, method, path, params=None, data=None):
+    def basic_request(self, method, path, params=None, data=None, headers=None):
         request_params = {
             'sessionId': self.session_id,
             'countryCode': self.country_code,
@@ -118,10 +318,21 @@ class Session(object):
         }
         if params:
             request_params.update(params)
+
+        if not headers:
+            headers = {}
+        if self.token_type:
+            headers['authorization'] = self.token_type + ' ' + self.access_token
         url = urljoin(self._config.api_location, path)
-        request = requests.request(method, url, params=request_params, data=data)
+        request = requests.request(method, url, params=request_params, data=data, headers=headers)
+        return request
+
+    def request(self, method, path, params=None, data=None, headers=None):
+        request = self.basic_request(method, path, params, data, headers)
         log.debug("request: %s", request.request.url)
-        request.raise_for_status()
+        if not request.ok:
+            print(request.text)
+            request.raise_for_status()
         if request.content:
             log.debug("response: %s", json.dumps(request.json(), indent=4))
         return request
